import 'dart:async';

import 'package:flutter/widgets.dart';
import 'package:lux/lux.dart';
import 'package:provider/provider.dart';
import 'package:provider/single_child_widget.dart';

import 'core.dart';
import 'stream_listener.dart';

typedef EventBusWidgetListener<B, S> = void Function(BuildContext context, B eventBus, S state);

class EventBusListener<B extends EventStreamableSource<S>, S> extends EventBusListenerBase<B, S>
    with LuxListenerSingleChildWidget {
  /// {@macro event_bus_listener}
  EventBusListener({
    Key? key,
    required EventBusWidgetListener<B, S> listener,
    B? eventBus,
    StreamListenerCondition<S>? listenWhen,
    Widget? child,
  }) : super(
          key: key,
          listener: listener,
          eventBus: eventBus,
          child: child,
          listenWhen: listenWhen,
        );
}

class EventBusListenerBase<B extends EventStreamableSource<S>, S> extends SingleChildStatefulWidget {
  /// {@macro event_bus_listener_base}
  const EventBusListenerBase({
    Key? key,
    required this.listener,
    this.eventBus,
    this.listenWhen,
    this.child,
  }) : super(key: key, child: child);

  final B? eventBus;

  final Widget? child;

  final EventBusWidgetListener<B, S> listener;

  final StreamListenerCondition<S>? listenWhen;

  @override
  SingleChildState<EventBusListenerBase<B, S>> createState() => _EventBusListenerBaseState<B, S>();
}

class _EventBusListenerBaseState<B extends EventStreamableSource<S>, S>
    extends SingleChildState<EventBusListenerBase<B, S>> {
  StreamSubscription<S>? _subscription;
  late B _eventBus;

  @override
  void initState() {
    super.initState();
    _eventBus = widget.eventBus ?? context.read<B>();
    _subscribe();
  }

  @override
  void didUpdateWidget(EventBusListenerBase<B, S> oldWidget) {
    super.didUpdateWidget(oldWidget);
    final oldEventBus = oldWidget.eventBus ?? context.read<B>();
    final currentEventBus = widget.eventBus ?? oldEventBus;
    if (oldEventBus != currentEventBus) {
      if (_subscription != null) {
        _unsubscribe();
        _eventBus = currentEventBus;
      }
      _subscribe();
    }
  }

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    final eventBus = widget.eventBus ?? context.read<B>();
    if (_eventBus != eventBus) {
      if (_subscription != null) {
        _unsubscribe();
        _eventBus = eventBus;
      }
      _subscribe();
    }
  }

  @override
  Widget buildWithChild(BuildContext context, Widget? child) {
    assert(child != null, '''${widget.runtimeType} used outside of EventBusListener must specify a child''');
    if (widget.eventBus == null) {
      context.select<B, bool>((eventBus) => identical(_eventBus, eventBus));
    }
    return child!;
  }

  @override
  void dispose() {
    _unsubscribe();
    super.dispose();
  }

  void _subscribe() {
    _subscription = _eventBus.stream.listen((state) {
      if (widget.listenWhen == null || widget.listenWhen!(state) == true) {
        widget.listener(context, _eventBus, state);
      }
    });
  }

  void _unsubscribe() {
    _subscription?.cancel();
    _subscription = null;
  }
}
