import 'package:flutter/widgets.dart';
import 'package:lux/lux.dart';
import 'package:provider/provider.dart';

import 'store_listener.dart';

/// Signature for the `builder` function which takes the `BuildContext` and
/// [state] and is responsible for returning a widget which is to be rendered.
/// This is analogous to the `builder` function in [StreamBuilder].
typedef StoreWidgetBuilder<B, S> = Widget Function(BuildContext context, B store, S state);

/// Signature for the `buildWhen` function which takes the previous `state` and
/// the current `state` and is responsible for returning a [bool] which
/// determines whether to rebuild [StoreBuilder] with the current `state`.
typedef StoreBuilderCondition<S> = bool Function(S previous, S current);

class StoreBuilder<B extends StateStreamable<S>, S> extends StoreBuilderBase<B, S> {
  /// {@macro store_builder}
  /// {@macro store_builder_build_when}
  const StoreBuilder({
    Key? key,
    required this.builder,
    B? store,
    StoreBuilderCondition<S>? buildWhen,
  }) : super(key: key, store: store, buildWhen: buildWhen);

  /// The [builder] function which will be invoked on each widget build.
  /// The [builder] takes the `BuildContext` and current `state` and
  /// must return a widget.
  /// This is analogous to the [builder] function in [StreamBuilder].
  final StoreWidgetBuilder<B, S> builder;

  @override
  Widget build(BuildContext context, B store, S state) => builder(context, store, state);
}

abstract class StoreBuilderBase<B extends StateStreamable<S>, S> extends StatefulWidget {
  /// {@macro store_builder_base}
  const StoreBuilderBase({Key? key, this.store, this.buildWhen}) : super(key: key);

  /// The [store] that the [StoreBuilderBase] will interact with.
  /// If omitted, [StoreBuilderBase] will automatically perform a lookup using
  /// [StoreProvider] and the current `BuildContext`.
  final B? store;

  /// {@macro store_builder_build_when}
  final StoreBuilderCondition<S>? buildWhen;

  /// Returns a widget based on the `BuildContext` and current [state].
  Widget build(BuildContext context, B store, S state);

  @override
  State<StoreBuilderBase<B, S>> createState() => _StoreBuilderBaseState<B, S>();
}

class _StoreBuilderBaseState<B extends StateStreamable<S>, S> extends State<StoreBuilderBase<B, S>> {
  late B _store;
  late S _state;

  @override
  void initState() {
    super.initState();
    _store = widget.store ?? context.read<B>();
    _state = _store.state;
  }

  @override
  void didUpdateWidget(StoreBuilderBase<B, S> oldWidget) {
    super.didUpdateWidget(oldWidget);
    final oldStore = oldWidget.store ?? context.read<B>();
    final currentStore = widget.store ?? oldStore;
    if (oldStore != currentStore) {
      _store = currentStore;
      _state = _store.state;
    }
  }

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    final store = widget.store ?? context.read<B>();
    if (_store != store) {
      _store = store;
      _state = _store.state;
    }
  }

  @override
  Widget build(BuildContext context) {
    if (widget.store == null) {
      context.select<B, bool>((store) => identical(_store, store));
    }
    return StoreListener<B, S>(
      store: _store,
      listenWhen: widget.buildWhen,
      listener: (context, store, state) => setState(() => _state = state),
      child: widget.build(context, _store, _state),
    );
  }
}
