import 'package:flutter/widgets.dart';
import 'package:lux/lux.dart';
import 'package:provider/provider.dart';

import 'store_builder.dart';
import 'store_listener.dart';

class StoreConsumer<B extends StateStreamable<S>, S> extends StatefulWidget {
  /// {@macro store_consumer}
  const StoreConsumer({
    Key? key,
    required this.builder,
    required this.listener,
    this.store,
    this.buildWhen,
    this.listenWhen,
  }) : super(key: key);

  /// The [store] that the [StoreConsumer] will interact with.
  /// If omitted, [StoreConsumer] will automatically perform a lookup using
  /// `StoreProvider` and the current `BuildContext`.
  final B? store;

  /// The [builder] function which will be invoked on each widget build.
  /// The [builder] takes the `BuildContext` and current `state` and
  /// must return a widget.
  /// This is analogous to the [builder] function in [StreamBuilder].
  final StoreWidgetBuilder<B, S> builder;

  /// Takes the `BuildContext` along with the [store] `state`
  /// and is responsible for executing in response to `state` changes.
  final StoreWidgetListener<B, S> listener;

  /// Takes the previous `state` and the current `state` and is responsible for
  /// returning a [bool] which determines whether or not to trigger
  /// [builder] with the current `state`.
  final StoreBuilderCondition<S>? buildWhen;

  /// Takes the previous `state` and the current `state` and is responsible for
  /// returning a [bool] which determines whether or not to call [listener] of
  /// [StoreConsumer] with the current `state`.
  final StoreListenerCondition<S>? listenWhen;

  @override
  State<StoreConsumer<B, S>> createState() => _StoreConsumerState<B, S>();
}

class _StoreConsumerState<B extends StateStreamable<S>, S> extends State<StoreConsumer<B, S>> {
  late B _store;

  @override
  void initState() {
    super.initState();
    _store = widget.store ?? context.read<B>();
  }

  @override
  void didUpdateWidget(StoreConsumer<B, S> oldWidget) {
    super.didUpdateWidget(oldWidget);
    final oldStore = oldWidget.store ?? context.read<B>();
    final currentStore = widget.store ?? oldStore;
    if (oldStore != currentStore) _store = currentStore;
  }

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    final store = widget.store ?? context.read<B>();
    if (_store != store) _store = store;
  }

  @override
  Widget build(BuildContext context) {
    if (widget.store == null) {
      context.select<B, bool>((store) => identical(_store, store));
    }
    return StoreBuilder<B, S>(
      store: _store,
      builder: widget.builder,
      buildWhen: (previous, current) {
        if (widget.listenWhen?.call(previous, current) ?? true) {
          widget.listener(context, _store, current);
        }
        return widget.buildWhen?.call(previous, current) ?? true;
      },
    );
  }
}
