import 'dart:async';

import 'package:flutter/widgets.dart';
import 'package:lux/lux.dart';
import 'package:provider/provider.dart';
import 'package:provider/single_child_widget.dart';

import 'core.dart';

/// Signature for the `listener` function which takes the `BuildContext` along
/// with the `state` and is responsible for executing in response to
/// `state` changes.
typedef StoreWidgetListener<B, S> = void Function(BuildContext context, B store, S state);

/// Signature for the `listenWhen` function which takes the previous `state`
/// and the current `state` and is responsible for returning a [bool] which
/// determines whether or not to call [StoreWidgetListener] of [StoreListener]
/// with the current `state`.
typedef StoreListenerCondition<S> = bool Function(S previous, S current);

class StoreListener<B extends StateStreamable<S>, S> extends StoreListenerBase<B, S> with LuxListenerSingleChildWidget {
  /// {@macro store_listener}
  /// {@macro store_listener_listen_when}
  const StoreListener({
    Key? key,
    required StoreWidgetListener<B, S> listener,
    B? store,
    StoreListenerCondition<S>? listenWhen,
    Widget? child,
  }) : super(
          key: key,
          child: child,
          listener: listener,
          store: store,
          listenWhen: listenWhen,
        );
}

abstract class StoreListenerBase<B extends StateStreamable<S>, S> extends SingleChildStatefulWidget {
  /// {@macro store_listener_base}
  const StoreListenerBase({
    Key? key,
    required this.listener,
    this.store,
    this.child,
    this.listenWhen,
  }) : super(key: key, child: child);

  /// The widget which will be rendered as a descendant of the
  /// [StoreListenerBase].
  final Widget? child;

  /// The [store] whose `state` will be listened to.
  /// Whenever the [store]'s `state` changes, [listener] will be invoked.
  final B? store;

  /// The [StoreWidgetListener] which will be called on every `state` change.
  /// This [listener] should be used for any code which needs to execute
  /// in response to a `state` change.
  final StoreWidgetListener<B, S> listener;

  /// {@macro store_listener_listen_when}
  final StoreListenerCondition<S>? listenWhen;

  @override
  SingleChildState<StoreListenerBase<B, S>> createState() => _StoreListenerBaseState<B, S>();
}

class _StoreListenerBaseState<B extends StateStreamable<S>, S> extends SingleChildState<StoreListenerBase<B, S>> {
  StreamSubscription<S>? _subscription;
  late B _store;
  late S _previousState;

  @override
  void initState() {
    super.initState();
    _store = widget.store ?? context.read<B>();
    _previousState = _store.state;
    _subscribe();
  }

  @override
  void didUpdateWidget(StoreListenerBase<B, S> oldWidget) {
    super.didUpdateWidget(oldWidget);
    final oldStore = oldWidget.store ?? context.read<B>();
    final currentStore = widget.store ?? oldStore;
    if (oldStore != currentStore) {
      if (_subscription != null) {
        _unsubscribe();
        _store = currentStore;
        _previousState = _store.state;
      }
      _subscribe();
    }
  }

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    final store = widget.store ?? context.read<B>();
    if (_store != store) {
      if (_subscription != null) {
        _unsubscribe();
        _store = store;
        _previousState = _store.state;
      }
      _subscribe();
    }
  }

  @override
  Widget buildWithChild(BuildContext context, Widget? child) {
    assert(
      child != null,
      '''${widget.runtimeType} used outside of LuxMultiListener must specify a child''',
    );
    if (widget.store == null) {
      context.select<B, bool>((store) => identical(_store, store));
    }
    return child!;
  }

  @override
  void dispose() {
    _unsubscribe();
    super.dispose();
  }

  void _subscribe() {
    _subscription = _store.stream.listen((state) {
      if (widget.listenWhen?.call(_previousState, state) ?? true) {
        widget.listener(context, _store, state);
      }
      _previousState = state;
    });
  }

  void _unsubscribe() {
    _subscription?.cancel();
    _subscription = null;
  }
}
