import 'package:flutter/widgets.dart';
import 'package:lux/lux.dart';
import 'package:provider/provider.dart';

import 'store_builder.dart';
import 'store_listener.dart';

/// Signature for the `selector` function which
/// is responsible for returning a selected value, [T], based on [state].
typedef StoreWidgetSelector<S, T> = T Function(S state);

class StoreSelector<B extends StateStreamable<S>, S, T> extends StatefulWidget {
  /// {@macro store_selector}
  const StoreSelector({
    Key? key,
    required this.selector,
    required this.builder,
    this.store,
  }) : super(key: key);

  /// The [store] that the [StoreSelector] will interact with.
  /// If omitted, [StoreSelector] will automatically perform a lookup using
  /// [StoreProvider] and the current [BuildContext].
  final B? store;

  /// The [builder] function which will be invoked
  /// when the selected state changes.
  /// The [builder] takes the [BuildContext] and selected `state` and
  /// must return a widget.
  /// This is analogous to the [builder] function in [StoreBuilder].
  final StoreWidgetBuilder<B, T> builder;

  /// The [selector] function which will be invoked on each widget build
  /// and is responsible for returning a selected value of type [T] based on
  /// the current state.
  final StoreWidgetSelector<S, T> selector;

  @override
  State<StoreSelector<B, S, T>> createState() => _StoreSelectorState<B, S, T>();
}

class _StoreSelectorState<B extends StateStreamable<S>, S, T> extends State<StoreSelector<B, S, T>> {
  late B _store;
  late T _state;

  @override
  void initState() {
    super.initState();
    _store = widget.store ?? context.read<B>();
    _state = widget.selector(_store.state);
  }

  @override
  void didUpdateWidget(StoreSelector<B, S, T> oldWidget) {
    super.didUpdateWidget(oldWidget);
    final oldStore = oldWidget.store ?? context.read<B>();
    final currentStore = widget.store ?? oldStore;
    if (oldStore != currentStore) {
      _store = currentStore;
      _state = widget.selector(_store.state);
    }
  }

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    final store = widget.store ?? context.read<B>();
    if (_store != store) {
      _store = store;
      _state = widget.selector(_store.state);
    }
  }

  @override
  Widget build(BuildContext context) {
    if (widget.store == null) {
      context.select<B, bool>((store) => identical(_store, store));
    }
    return StoreListener<B, S>(
      store: _store,
      listener: (context, store, state) {
        final selectedState = widget.selector(state);
        if (_state != selectedState) setState(() => _state = selectedState);
      },
      child: widget.builder(context, _store, _state),
    );
  }
}
