import 'package:flutter/widgets.dart';
import 'package:lux/lux.dart';
import 'package:provider/provider.dart';

import 'store_listener.dart';

typedef StoreWidgetWatcher<S, T> = T Function(S state);

typedef StoreWidgetListStateBuilder<B, T> = Widget Function(BuildContext context, B store, T state);

class StoreWatcher<B extends StateStreamable<S>, S, T> extends StatefulWidget {
  /// {@macro store_watcher}
  const StoreWatcher({
    Key? key,
    required this.watcher,
    required this.builder,
    this.store,
  }) : super(key: key);

  final B? store;

  final StoreWidgetListStateBuilder<B, T> builder;

  final StoreWidgetWatcher<S, T> watcher;

  @override
  State<StoreWatcher<B, S, T>> createState() => _StoreWatcherState<B, S, T>();
}

class _StoreWatcherState<B extends StateStreamable<S>, S, T> extends State<StoreWatcher<B, S, T>> {
  late B _store;
  late T _state;

  @override
  void initState() {
    super.initState();
    _store = widget.store ?? context.read<B>();
    _state = widget.watcher(_store.state);
  }

  @override
  void didUpdateWidget(StoreWatcher<B, S, T> oldWidget) {
    super.didUpdateWidget(oldWidget);
    final oldStore = oldWidget.store ?? context.read<B>();
    final currentStore = widget.store ?? oldStore;
    if (oldStore != currentStore) {
      _store = currentStore;
      _state = widget.watcher(_store.state);
    }
  }

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    final store = widget.store ?? context.read<B>();
    if (_store != store) {
      _store = store;
      _state = widget.watcher(_store.state);
    }
  }

  bool _compareWith(T selected) {
    if (selected is Iterable && _state is Iterable) {
      Iterable state = _state as Iterable;
      if (selected.length != state.length) {
        return false;
      }
      for (int i = 0, len = selected.length; i < len; i++) {
        if (selected.elementAt(i) != state.elementAt(i)) {
          return false;
        }
      }
    } else if (selected is Map && _state is Map) {
      Map state = _state as Map;
      if (selected.length != state.length) {
        return false;
      }
      for (var entry in selected.entries) {
        if (!state.containsKey(entry.key)) {
          return false;
        }
        if (entry.value != state[entry.key]) {
          return false;
        }
      }
    } else {
      return selected == _state;
    }
    return true;
  }

  @override
  Widget build(BuildContext context) {
    if (widget.store == null) {
      context.select<B, bool>((store) => identical(_store, store));
    }
    return StoreListener<B, S>(
      store: _store,
      listener: (context, store, state) {
        final selectedValue = widget.watcher(state);
        if (!_compareWith(selectedValue)) {
          setState(() {
            _state = selectedValue;
          });
        }
      },
      child: widget.builder(context, _store, _state),
    );
  }
}
