import 'dart:async';

import 'package:flutter/widgets.dart';
import 'package:provider/single_child_widget.dart';

import 'core.dart';

typedef StreamWidgetListener<S> = void Function(BuildContext context, S state);

typedef StreamListenerCondition<S> = bool Function(S state);

class StreamListener<B extends Stream<S>, S> extends StreamListenerBase<B, S> with LuxListenerSingleChildWidget {
  /// {@macro steam_listener}
  const StreamListener({
    Key? key,
    required StreamWidgetListener<S> listener,
    required B stream,
    StreamListenerCondition<S>? listenWhen,
    Widget? child,
  }) : super(
          key: key,
          listener: listener,
          stream: stream,
          child: child,
          listenWhen: listenWhen,
        );
}

class StreamListenerBase<B extends Stream<S>, S> extends SingleChildStatefulWidget {
  /// {@macro steam_listener_base}
  const StreamListenerBase({
    Key? key,
    required this.listener,
    required this.stream,
    this.listenWhen,
    this.child,
  }) : super(key: key, child: child);

  final B stream;

  final Widget? child;

  final StreamWidgetListener<S> listener;

  final StreamListenerCondition<S>? listenWhen;

  @override
  SingleChildState<StreamListenerBase<B, S>> createState() => _StreamListenerBaseState<B, S>();
}

class _StreamListenerBaseState<B extends Stream<S>, S> extends SingleChildState<StreamListenerBase<B, S>> {
  StreamSubscription<S>? _subscription;
  late B _stream;

  @override
  void initState() {
    super.initState();
    _stream = widget.stream;
    _subscribe();
  }

  @override
  void didUpdateWidget(StreamListenerBase<B, S> oldWidget) {
    super.didUpdateWidget(oldWidget);
    final oldStream = oldWidget.stream;
    final currentStream = widget.stream;
    if (oldStream != currentStream) {
      if (_subscription != null) {
        _unsubscribe();
        _stream = currentStream;
      }
      _subscribe();
    }
  }

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    final stream = widget.stream;
    if (_stream != stream) {
      if (_subscription != null) {
        _unsubscribe();
        _stream = stream;
      }
      _subscribe();
    }
  }

  @override
  Widget buildWithChild(BuildContext context, Widget? child) {
    assert(child != null, '''${widget.runtimeType} used outside of StreamListener must specify a child''');
    return child!;
  }

  @override
  void dispose() {
    _unsubscribe();
    super.dispose();
  }

  void _subscribe() {
    _subscription = _stream.listen((state) {
      if (widget.listenWhen == null || widget.listenWhen!(state) == true) {
        widget.listener(context, state);
      }
    });
  }

  void _unsubscribe() {
    _subscription?.cancel();
    _subscription = null;
  }
}
